/*!
 * \file
 * \brief   Main application file for PDS HW2 (MPI)
 *
 * \author
 *    Christos Choutouridis AEM:8997
 *    <cchoutou@ece.auth.gr>
 */

#include <exception>
#include <iostream>

#include "utils.hpp"
#include "config.h"


// Global session data
session_t  session;
MPI_t<>    mpi;


/*!
 * A small command line argument parser
 * \return  The status of the operation
 */
bool get_options(int argc, char* argv[]){
    bool status =true;

    // iterate over the passed arguments
    for (int i=1 ; i<argc ; ++i) {
        std::string arg(argv[i]);     // get current argument

        if (arg == "-x" || arg == "--xxxxx") {
            if (i+2 < argc) {
//                session.corpusMtxFile = std::string(argv[++i]);
//                session.corpusDataSet = std::string(argv[++i]);
            }
            else
                status = false;
        }

        else if (arg == "-v" || arg == "--verbose")
            session.verbose = true;
        else if (arg == "-h" || arg == "--help") {
            std::cout << "distBitonic - A distributed bitonic sort\n\n";
            std::cout << "distBitonic -x <> [-v]\n";
            std::cout << '\n';
            std::cout << "Options:\n\n";
            std::cout << "   -v | --verbose\n";
            std::cout << "      Request a more verbose output to stdout.\n\n";
            std::cout << "   -h | --help\n";
            std::cout << "      Prints this and exit.\n\n";
            std::cout << "Examples:\n\n";
            std::cout << "   ...Example case...:\n";
            std::cout << "   > distBitonic -x <xxxxx>  \n\n";

            exit(0);
        }
        else {   // parse error
            std::cout << "Invocation error. Try -h for details.\n";
            status = false;
        }
    }

    return status;
}



#if !defined TESTING
int main(int argc, char* argv[]) try {
    // Initialize MPI environment
    mpi.init(&argc, &argv);

    // try to read command line (after MPI parsing)
    if (!get_options(argc, argv))
        exit(1);

    #if defined DEBUG
        /*
         * In case of a debug build we will wait here until sleep_wait
         * will reset via debugger. In order to do that the user must attach
         * debugger to all processes. For example:
         *  $> mpirun -np 2 ./<program path>
         *  $> ps aux | grep <program>
         *  $> gdb <program> <PID1>
         *  $> gdb <program> <PID2>
         */
    #if defined TESTING
    volatile bool sleep_wait = false;
    #else
    volatile bool sleep_wait = true;
    #endif
    while (sleep_wait)
        sleep(1);
    #endif

    // Print off a hello world message
//    std::cout   << "Hello world from processor: " << mpi.processor_name
//                << " rank " << mpi.world_rank
//                << " out of " << mpi.world_size << " processors\n";


//    distBitonic (2, Data);
//    distBitonic (4, Data);

    mpi.finalize();
    return 0;
}
catch (std::exception& e) {
    //we probably pollute the user's screen. Comment `cerr << ...` if you don't like it.
    std::cerr << "Error: " << e.what() << '\n';
    exit(1);
}

#else

#include <gtest/gtest.h>
#include <exception>

GTEST_API_ int main(int argc, char **argv) try {
   testing::InitGoogleTest(&argc, argv);
   return RUN_ALL_TESTS();
}
catch (std::exception& e) {
    std::cout << "Exception: " << e.what() << '\n';
}

#endif